-----------------------------------------------------------------------------

local SML = AceLibrary("SharedMedia-1.0")
local SML_statusbar = SML:List("statusbar")
local SML_font = SML:List("font")

local Dewdrop = AceLibrary:HasInstance("Dewdrop-2.0") and AceLibrary("Dewdrop-2.0")
local AssessmentUserInterface = Assessment:GetModule("UserInterface")

-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentUserInterface")
L:RegisterTranslations("enUS", function() return {
	["Windows"] = true,
	["Window options."] = true,
	["Display Sets"] = true,
	["Display set options."] = true,
	["Add New Window"] = true,
	["Add a new assessment window to the user interface."] = true,

	["Window %d"] = true,
	["Options for assessment window %d"] = true,
	["Display Set"] = true,
	["Choose what to show for this window"] = true,
	["Visible"] = true,
	["Make the assessment window visible"] = true,

	["Lock"] = true,
	["Enable to prevent this window from being draggable."] = true,

	["Update Rate"] = true,
	["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."] = true,

	["Backdrop Color"] = true,
	["Change the backdrop color and alpha for the window."] = true,
	["Border Style"] = true,
	["Change the border style for the window."] = true,
	["None"] = true,
	["Default"] = true,
	["Textured"] = true,
	["Square"] = true,
	["Border Color"] = true,
	["Change the border color and alpha for the window."] = true,

	["Bar Options"] = true,
	["Bar options for this window"] = true,
	["Bar Count"] = true,
	["Number of bars to show"] = true,
	["Texture"] = true, 
	["Sets the bar texture"] = true,

	["Title Options"] = true,
	["Title options."] = true,
	["Enable"] = true,
	["Enable title text."] = true,
	["Alignment"] = true,
	["Select how to align the title."] = true,
	["Font"] = true,
	["Font for title."] = true,
	["Font Size"] = true,
	["Font size for title."] = true,
	["Font Shadow"] = true,
	["Show font shadow on title text"] = true,
	["Outline"] = true,
	["Sets the outline for title text"] = true,
	["Title Height"] = true,
	["Title height."] = true,
	["Background Color"] = true,
	["Change the title backdrop color."] = true,
	["Text Color"] = true,
	["Change the title text color."] = true,

	["Remove"] = true,
	["Remove this window from display"] = true,

	["Add New Display Set"] = true,
	["Adds a new display set to the user interface."] = true,
	["Enter a new display set name"] = true,

	["Options for display set %s"] = true,

	["List Type"] = true,
	["Select a list type that controls what is displayed."] = true,
	["Bar Value Tag"] = true,
	["Select how to calculate bar values."] = true,
	["Sort Key Tag"] = true,
	["Select how to sort the bars."] = true,
	["Descending Sort"] = true,
	["Enable to use descending sort. Disable to use ascending."] = true,

	["Favorite"] = true,
	["Show this display set as an option when clicking on a window."] = true,
	
	["Title Tag"] = true,
	["Select what to show for the window's title."] = true,

	["Filter Type"] = true,
	["Select how to filter the list."] = true,

	["Fight Data"] = true,
	["Select which fight data set to display."] = true,
	["Total Data"] = true,
	["Current Fight"] = true,
	["Previous Fight"] = true,

	["Text"] = true,
	["Text options."] = true,
	["Add New Text"] = true,
	["Adds a new text to the display set."] = true,

	["Text %d"] = true,
	["Options for text %d"] = true,

	["Text Tag"] = true,
	["Select how to display text."] = true,

	["Alpha"] = true,
	["Set the text alpha."] = true,

	["Alignment"] = true,
	["Select how to align the text."] = true,
	["Left"] = true,
	["Center"] = true,
	["Right"] = true,

	["Font"] = true,
	["Font for text."] = true,

	["Font Size"] = true,
	["Font size for text."] = true,

	["Font Shadow"] = true,
	["Show font shadow on text"] = true,
	["Outline"] = true,
	["Sets the outline for text"] = true,

	["Remove"] = true,
	["Remove this text."] = true,

	["Remove"] = true,
	["Remove this display set."] = true,
	["This display set cannot be removed currently because one or more windows require its data."] = true,
	
	["Any Tag String"] = true,
	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,
} end)

-----------------------------------------------------------------------------

Assessment.Options.args.uiSpacer =
{
	type = 'header',
	order = 50,
}

Assessment.Options.args.windows =
{
	type = "group",
	name = L["Windows"],
	desc = L["Window options."],
	order = 60,
	args = { },
}

local windowsArgs =
{ 
	add =
	{
		type = "execute",
		name = L["Add New Window"],
		desc = L["Add a new assessment window to the user interface."],
		func = function() AssessmentUserInterface:AddWindow() end,
		order = -1
	},
}

Assessment.Options.args.displaySets =
{
	type = "group",
	name = L["Display Sets"],
	desc = L["Display set options."],
	order = 60,
	args = { },
}

local displaySetsArgs =
{
	add =
	{
		type = "execute",
		name = L["Add New Display Set"],
		desc = L["Adds a new display set to the user interface."],
		func = function() AssessmentUserInterface:AddDisplaySet() end,
		order = -1
	},
}

-----------------------------------------------------------------------------

local function windowGet(passValue)
	local s = ("return function(t) return t[%d].%s end"):format(unpack(passValue))
	return loadstring(s)()(AssessmentUserInterface.db.profile.windows)
end

local function windowSet(passValue, value)
	local s = ("return function(t, v) t[%d].%s = v end"):format(unpack(passValue))
	loadstring(s)()(AssessmentUserInterface.db.profile.windows, value)
	AssessmentUserInterface:UpdateWindowOptions(passValue[1])
end

local function windowGetDisplaySet(passValue)
	local windowOptions = AssessmentUserInterface.db.profile.windows[passValue[1]]
	return windowOptions.displaySetStack[windowOptions.displaySetStackLevel]
end

local function windowSetDisplaySet(passValue, value)
	local windowOptions = AssessmentUserInterface.db.profile.windows[passValue[1]]
	
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = value

	AssessmentUserInterface:UpdateWindowOptions(passValue[1])
end

local function displaySetGet(passValue)
	local s = ("return function(t) return t[\"%s\"].%s end"):format(unpack(passValue))
	local value = loadstring(s)()(AssessmentUserInterface.db.profile.displaySets)
	if passValue[2] == "fightIndex" then
		value = tostring(value)
	end
	return value
end

local function displaySetSet(passValue, value)
	if passValue[2] == "fightIndex" then
		value = tonumber(value)
	end

	local s = ("return function(t, v) t[\"%s\"].%s = v end"):format(unpack(passValue))
	loadstring(s)()(AssessmentUserInterface.db.profile.displaySets, value)
	AssessmentUserInterface:UpdateDisplaySets(passValue[1])
end

local function textGet(passValue)
	local s = ("return function(t) return t[\"%s\"].text[%d].%s end"):format(passValue[1][1], passValue[1][2], passValue[2])
	return loadstring(s)()(AssessmentUserInterface.db.profile.displaySets)
end

local function textSet(passValue, value)
	local s = ("return function(t, v) t[\"%s\"].text[%d].%s = v end"):format(passValue[1][1], passValue[1][2], passValue[2])
	loadstring(s)()(AssessmentUserInterface.db.profile.displaySets, value)
	AssessmentUserInterface:UpdateDisplaySets(passValue[1][1])
end

local function windowRemove(passValue)
	AssessmentUserInterface:RemoveWindow(passValue[1])
end

local function displaySetRemove(passValue)
	AssessmentUserInterface:RemoveDisplaySet(passValue[1])
end

-----------------------------------------------------------------------------

local displaySetValidate = { }
local listTypeValidate	 = { }
local filterTypeValidate = { }

-----------------------------------------------------------------------------

local windowOptionsTemplate =
{
	type = "group",
--	name = ["Window %d"] = true,
--	desc = ["Options for assessment window %d"] = true,
	args =
	{
		visible =
		{
			type = "toggle",
			name = L["Visible"],
			desc = L["Make the assessment window visible"],
			get = windowGet,
			set = windowSet,
			passValue = "frame.visible",
			order = 1,
		},
		displaySet = 
		{
			type = "text",
			name = L["Display Set"],
			desc = L["Choose what to show for this window"],
			get = windowGetDisplaySet,
			set = windowSetDisplaySet,
			passValue = "displaySet",
			validate = displaySetValidate,
			order = 2,
		},
		lock =
		{
			type = "toggle",
			name = L["Lock"],
			desc = L["Enable to prevent this window from being draggable."],
			get = windowGet,
			set = windowSet,
			passValue = "frame.lock",
		},
		backdropColor = 
		{
			type = "color",
			name = L["Backdrop Color"],
			desc = L["Change the backdrop color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.backdropColor",
			hasAlpha = true,
		},
		borderStyle = 
		{
			type = "text",
			name = L["Border Style"],
			desc = L["Change the border style for the window."],
			get = windowGet,
			set = windowSet,
			validate = { None = L["None"], Default = L["Default"], Textured = L["Textured"], Square = L["Square"] },
			passValue = "frame.borderStyle",
		},
		borderColor = 
		{
			type = "color",
			name = L["Border Color"],
			desc = L["Change the border color and alpha for the window."],
			get = function(passValue) return unpack(windowGet(passValue)) end,
			set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
			passValue = "frame.borderColor",
			hasAlpha = true,
		},
		updateRate =
		{
			type = "range",
			name = L["Update Rate"],
			desc = L["Controls the maximum number of times per second to update this window. Use 0 for continuous updates."],
			min  = 0,
			max  = 60,
			step = 1,
			passValue = "updateRate",
			get = windowGet,
			set = windowSet,
		},
		bars =
		{
			type = "group",
			name = L["Bar Options"],
			desc = L["Bar options for this window"],
			args =
			{
				count =
				{
					type = "range",
					name = L["Bar Count"],
					desc = L["Number of bars to show"],
					min	 = 1,
					max  = 50,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "bar.count",
					order = 1,
				},
				texture =
				{
					type = "text", 
					name = L["Texture"], 
					desc = L["Sets the bar texture"],
					get = windowGet,
					set = windowSet,
					validate = SML_statusbar,
					passValue = "bar.texture",
					order = 2,
				},
			},
		},
		title =
		{
			type = "group",
			name = L["Title Options"],
			desc = L["Title options."],
			args =
			{
				enable = 
				{
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable title text."],
					get = windowGet,
					set = windowSet,
					passValue = "title.enable",
					order = 1,
				},
				alignment =
				{
					type = "text",
					name = L["Alignment"],
					desc = L["Select how to align the title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.alignment",
					validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
				},
				font =
				{
					type = "text",
					name = L["Font"],
					desc = L["Font for title."],
					get = windowGet,
					set = windowSet,
					passValue = "title.font",
					validate = SML_font,
				},
				height =
				{
					type = "range",
					name = L["Title Height"],
					desc = L["Title height."],
					min  = 0,
					max  = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.height",
				},
				fontSize =
				{
					type = "range",
					name = L["Font Size"],
					desc = L["Font size for title."],
					min = 4,
					max = 32,
					step = 1,
					get = windowGet,
					set = windowSet,
					passValue = "title.fontSize",
				},
				shadow =
				{
					type = "toggle",
					name = L["Font Shadow"],
					desc = L["Show font shadow on title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.shadow",
				},
				backgroundColor = 
				{
					type = "color",
					name = L["Background Color"],
					desc = L["Change the title backdrop color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.backgroundColor",
					hasAlpha = true,
				},
				textColor = 
				{
					type = "color",
					name = L["Text Color"],
					desc = L["Change the title text color."],
					get = function(passValue) return unpack(windowGet(passValue)) end,
					set = function(passValue, r, g, b, a) windowSet(passValue, { r, g, b, a }) end,
					passValue = "title.textColor",
					hasAlpha = true,
				},
				outline =
				{
					type = "text",
					name = L["Outline"],
					desc = L["Sets the outline for title text"],
					get = windowGet,
					set = windowSet,
					passValue = "title.outline",
					validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
				},
			},
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this window from display"],
			func = windowRemove,
			passValue = false,
			order = -1,
		},
	}
}

local displaySetOptionsTemplate =
{
	type = "group",
--	name = ["%s"] = true,
--	desc = ["Options for display set %s"] = true,
	args =
	{
		listType =
		{
			type = "text",
			name = L["List Type"],
			desc = L["Select a list type that controls what is displayed."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "listType",
			validate = listTypeValidate,
			order = 1,
		},
		barValueTag =
		{
			type = "text",
			name = L["Bar Value Tag"],
			desc = L["Select how to calculate bar values."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "barValueTag",
			usage = L["Any Tag String"],
		},
		sortKeyTag =
		{
			type = "text",
			name = L["Sort Key Tag"],
			desc = L["Select how to sort the bars."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "sortKeyTag",
			usage = L["Any Tag String"],
		},
		titleTag =
		{
			type = "text",
			name = L["Title Tag"],
			desc = L["Select what to show for the window's title."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "titleTag",
			usage = L["Any Tag String"],
		},
		filterType =
		{
			type = "text",
			name = L["Filter Type"],
			desc = L["Select how to filter the list."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "filterType",
			validate = filterTypeValidate,
		},
		descendingSort =
		{
			type = "toggle",
			name = L["Descending Sort"],
			desc = L["Enable to use descending sort. Disable to use ascending."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "descendingSort",
		},
		fightData =
		{
			type = "text",
			name = L["Fight Data"],
			desc = L["Select which fight data set to display."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "fightIndex",
			validate = { ["0"] = L["Total Data"], ["1"] = L["Current Fight"], ["2"] = L["Previous Fight"] },
		},
		text =
		{
			type = "group",
			name = L["Text"],
			desc = L["Text options."],
			args =
			{
				add =
				{
					type = "execute",
					name = L["Add New Text"],
					desc = L["Adds a new text to the display set."],
					func = function(passValue) AssessmentUserInterface:AddText(passValue[1]) end,
					passValue = true,
					order = -1
				},
			},
		},

		favorite =
		{
			type = "toggle",
			name = L["Favorite"],
			desc = L["Show this display set as an option when clicking on a window."],
			get = displaySetGet,
			set = displaySetSet,
			passValue = "favorite",
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this display set."],
			func = displaySetRemove,
			passValue = false,
			order = -1,
		},
	}
}

local textOptionsTemplate =
{
	type = "group",
--	name = ["Text %d"] = true,
--	desc = ["Options for text %d"] = true,
	args =
	{
		tag =
		{
			type = "text",
			name = L["Text Tag"],
			desc = L["Select how to display text."],
			get = textGet,
			set = textSet,
			passValue = "tag",
			usage = L["Any Tag String"],
		},
		alignment =
		{
			type = "text",
			name = L["Alignment"],
			desc = L["Select how to align the text."],
			get = textGet,
			set = textSet,
			passValue = "alignment",
			validate = { LEFT = L["Left"], CENTER = L["Center"], RIGHT = L["Right"] },
		},
		alpha =
		{
			type = "range",
			name = L["Alpha"],
			desc = L["Set the text alpha."],
			min = 0,
			max = 1,
			step = 0.01,
			get = textGet,
			set = textSet,
			passValue = "alpha",
		},
		font =
		{
			type = "text",
			name = L["Font"],
			desc = L["Font for text."],
			get = textGet,
			set = textSet,
			passValue = "font",
			validate = SML_font,
		},
		fontSize =
		{
			type = "range",
			name = L["Font Size"],
			desc = L["Font size for text."],
			min = 4,
			max = 32,
			step = 1,
			get = textGet,
			set = textSet,
			passValue = "fontSize",
		},
		shadow =
		{
			type = "toggle",
			name = L["Font Shadow"],
			desc = L["Show font shadow on text"],
			get = textGet,
			set = textSet,
			passValue = "shadow",
		},
		outline =
		{
			type = "text",
			name = L["Outline"],
			desc = L["Sets the outline for text"],
			get = textGet,
			set = textSet,
			passValue = "outline",
			validate = { [""] = L["None"], OUTLINE = L["Normal"], THICKOUTLINE = L["Thick"] },
		},
		remove =
		{
			type = "execute",
			name = L["Remove"],
			desc = L["Remove this text."],
			func = function(passValue) AssessmentUserInterface:RemoveText(unpack(passValue[1])) end,
			passValue = false,
			order = -1,
		},
	}
}

-----------------------------------------------------------------------------

function AssessmentUserInterface:CreateValidateLists()
    for k,v in pairs(Assessment.ListType) do
        listTypeValidate[k] = v.description
    end 
            
    for k,v in pairs(Assessment.FilterType) do
        filterTypeValidate[k] = v.description
    end     
end

-----------------------------------------------------------------------------

local function copyOptionsTemplate(source, i)
	local result = Assessment:NewTable()

	for k,v in pairs(source) do
		if type(v) == "table" then
			result[k] = copyOptionsTemplate(v, i)
		else
			if k == "passValue" then
				result[k] = { i, v }
			else
				result[k] = v
			end
		end
	end
	
	return result
end

function AssessmentUserInterface:CreateOptions()
	for k in pairs(displaySetValidate) do
		displaySetValidate[k] = nil
	end

	for k in pairs(self.db.profile.displaySets) do
		displaySetValidate[k] = k
	end

	self:ReleaseTable(Assessment.Options.args.windows.args, true)
	Assessment.Options.args.windows.args = copyOptionsTemplate(windowsArgs)

	for i = 1, self.db.profile.windows.count do
		Assessment.Options.args.windows.args[tostring(i)] = self:CreateWindowOptions(i)
	end

	self:ReleaseTable(Assessment.Options.args.displaySets.args, true)
	Assessment.Options.args.displaySets.args = copyOptionsTemplate(displaySetsArgs)
	
	for k in pairs(self.db.profile.displaySets) do
		Assessment.Options.args.displaySets.args[k] = self:CreateDisplaySetOptions(k)
	end
end

function AssessmentUserInterface:CreateWindowOptions(i)
	local options = copyOptionsTemplate(windowOptionsTemplate, i)
	options.name = L["Window %d"]:format(i)
	options.desc = L["Options for assessment window %d"]:format(i)
	return options
end

function AssessmentUserInterface:CreateDisplaySetOptions(displaySetName)
	local options = copyOptionsTemplate(displaySetOptionsTemplate, displaySetName)
	options.name = displaySetName
	options.desc = L["Options for display set %s"]:format(displaySetName)

	for i = 1, self.db.profile.displaySets[displaySetName].text.count do
		-- We make it a string, otherwise AceConsole can't use it.....
		options.args.text.args[tostring(i)] = self:CreateTextOptions( { displaySetName, i } )
	end
	
	return options
end

function AssessmentUserInterface:CreateTextOptions(passValueData)
	local options = copyOptionsTemplate(textOptionsTemplate, passValueData)
	options.name = L["Text %d"]:format(passValueData[2])
	options.desc = L["Options for text %d"]:format(passValueData[2])
	return options
end

function AssessmentUserInterface:AddWindow()
	local windowOptions = self.db.profile.windows
	windowOptions.count = windowOptions.count + 1
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentUserInterface:RemoveWindow(index)
	local windowOptions = self.db.profile.windows
	local windows = self.windows

	local oldWindow = windows[index]
	local oldWindowOptions = windowOptions[index]

	for i = index, windowOptions.count do
		windowOptions[i] = windowOptions[i+1]
		windows[i] = windows[i+1]
	end
	windows[windowOptions.count] = oldWindow
	windowOptions[windowOptions.count] = oldWindowOptions
	
	oldWindow:Hide()
	
	windowOptions.count = windowOptions.count - 1
	
	self:CreateOptions()
end

function AssessmentUserInterface:UpdateWindowOptions(index)
	self.windows[index]:SetWindowOptions(self.db.profile.windows[index])
end

-----------------------------------------------------------------------------

StaticPopupDialogs["ASSESSMENT_ADD_DISPLAY_SET"] = 
{
	text = L["Enter a new display set name"],
	button1 = ACCEPT,
	button2 = CANCEL,
	hasEditBox = 1,
	OnAccept = function()
		local text = getglobal(this:GetParent():GetName().."EditBox"):GetText()
		AssessmentUserInterface:AddDisplaySet(text)
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus()
	end,
	EditBoxOnEnterPressed = function()
		local text = getglobal(this:GetParent():GetName().."EditBox"):GetText()
		AssessmentUserInterface:AddDisplaySet(text)
		this:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1,
}

function AssessmentUserInterface:AddDisplaySet(displaySetName)
	if displaySetName then
		self.db.profile.displaySets[displaySetName] = copyOptionsTemplate(profileDefaults.displaySets["**"])
		self:CreateOptions()
	else
		if Dewdrop then Dewdrop:Close() end
		StaticPopup_Show("ASSESSMENT_ADD_DISPLAY_SET")
	end
end

function AssessmentUserInterface:UpdateDisplaySets(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if windowOptions.displaySetStack[windowOptions.displaySetStackLevel] == displaySetName then
			self.windows[i]:SetWindowOptions(windowOptions)
		end
	end
end

function AssessmentUserInterface:RemoveDisplaySet(displaySetName)
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		
		for j = 1, windowOptions.displaySetStackLevel do
			if windowOptions.displaySetStack[j] == displaySetName then
				Assessment:Print(L["This display set cannot be removed currently because one or more windows require its data."])
				return
			end
		end
	end
	
	self.db.profile.displaySets[displaySetName] = nil
	self:CreateOptions()
end

-----------------------------------------------------------------------------

function AssessmentUserInterface:AddText(displaySetName)
	local text = self.db.profile.displaySets[displaySetName].text
	text.count = text.count + 1

	self:UpdateDisplaySets(displaySetName)
	self:CreateOptions()
end

function AssessmentUserInterface:RemoveText(displaySetName, removeIndex)
	local text = self.db.profile.displaySets[displaySetName].text
	
	for i = removeIndex, text.count do
		text[i] = text[i+1]
	end
	
	text.count = text.count-1

	self:UpdateDisplaySets(displaySetName)

	self:CreateOptions()
end

-----------------------------------------------------------------------------
